#include <time.h>
#include <crtdbg.h>

#include "gamex-cleanmem.h"
#undef new
#undef delete
#undef malloc
#undef free

//#define DEBUG_OUTPUT_ALL_ALLOCS
#ifdef DEBUG_OUTPUT_ALL_ALLOCS
	static long totalAllocated = 0;
//#	define DEBUG_OUTPUT_ALL_ALLOCS_ONLY_ON_CTRL
#	ifdef DEBUG_OUTPUT_ALL_ALLOCS_ONLY_ON_CTRL
#		include <windows.h>
#	endif
#endif
//#define DEBUG_EXTENDED_CHECK_ALLOCS


#ifdef CLEANMEM_DISABLE

	// if cleanMem was requested to be disabled,
	// do nothing but supply 2 empty functions:
	void BeginCheckingUnfreed() {}
	void DumpUnfreed() {}
	bool checkLeakScope = false;
	bool customFreeEnabled = false;

#else // !CLEANMEM_DISABLE:


	// stuff for keeping track of allocations / deallocations
	struct AllocInfo
	{
		unsigned long address;
		unsigned long size;
		const char * file;
		unsigned short line;
		bool isArray;

		inline bool operator == (const AllocInfo & op) const { return address == op.address; }

		inline AllocInfo (unsigned long addr) { address = addr; }
		inline AllocInfo () {}
	};

	typedef std::vector<AllocInfo> AllocList;

	AllocList *allocList;

	const char * allocDeleteFile = NULL;
	int allocDeleteLine = -1;

	bool leakDebugFileOpened = false;

	// outputs string to log file
	void OutputDebugStr(char *debugStr/*, bool andToDebugger = true*/)
	{
		// clear the output file if it hasn't been opened yet
		if(!leakDebugFileOpened)
		{
			debug.DeleteAnyFile(CLEAN_MEM_LOG_FILE);
			leakDebugFileOpened = true;
		}

		FILE *out;

		if((out = fopen(CLEAN_MEM_LOG_FILE, "a")) == NULL)
			fprintf(stderr, "Cannot open output file \"%s\".\n", debugStr);
		else
		{
			fputs(debugStr, out);

			fflush(out);
			fclose(out);
		}

//		if(andToDebugger)
//			debug.Output("%s", debugStr);
	};

	bool checkLeakScope = false;
	bool customFreeEnabled = false;
	bool isDebug = true;


	void BeginCheckingUnfreed()
	{
		if(!checkLeakScope && isDebug)
		{
			if(!/*debug.IsRunningDebug()*/IsDebuggerPresent() && !(GetAsyncKeyState(VK_CONTROL) & 0x8000))
			{
				isDebug = false;
			}
			else
			{
				checkLeakScope = true;
				customFreeEnabled = true;
				atexit(DumpUnfreed); // account for all memory leaks as late as possible (because this atexit() is called in the earliest possible scope)
				allocList = new(AllocList);
			}
		}
	}

	// call BeginCheckingUnfreed as early as possible to avoid some false positives with memory leaks in dynamically initialized static global objects
	int preStartupCode(void);
	#pragma data_seg(".CRT$XIU") // initialization user segment of CRT
	static int (*initptr)(void) = preStartupCode;
	#pragma data_seg() // reset data segment
	int preStartupCode(void)
	{
	   if(initptr)
		   BeginCheckingUnfreed();
	   return 0;
	}


	static void strcatAfter(char * str, const char * cat, int minFirstSegLen)
	{
		int len = strlen(str);

		if(len < minFirstSegLen)
		{
			for(int i = len ; i < minFirstSegLen ; i++)
				str[i] = ' ';

			str[minFirstSegLen] = '\0';
		}

		strcat(str, cat);
	}

	// shows a summary of unfreed memory
	void DumpUnfreed()
	{
		const bool wasEmpty = !allocList || allocList->empty();

		// don't dump anything if there were no memory leaks AND there is no existing dump file
		if(wasEmpty)
		{
			if(!debug.DoesFileExist(CLEAN_MEM_LOG_FILE))
				return;
		}

		AllocList::iterator i;
		unsigned long long totalSize = 0;
		char buf[1024*8];

		// make a string with the time and date in it
		char timebuf [256];
		char* timebufPtr = timebuf;
		_strtime(timebufPtr);
		strcat(timebufPtr, " ");
		timebufPtr += strlen(timebufPtr);
		_strdate(timebufPtr);

		sprintf(buf, "\n-----------------------------------------------------------\n");			OutputDebugStr(buf);
		sprintf(buf, " Begin Dump of Memory Leaks:         (At %s)\n", timebuf);	OutputDebugStr(buf);
		sprintf(buf, "-----------------------------------------------------------\n\n");			OutputDebugStr(buf);

		if(!allocList || allocList->empty())
		{
			sprintf(buf, "No memory leaks detected!\n");
			OutputDebugStr(buf);
		}
		else
		{
			int maxLineNumLength = 0;
			int maxFileLength = 0;

			// calculate maximum filename and line number lengths for a pretty printout
			for(unsigned int i = 0 ; i < allocList->size() ; i++)
			{
				AllocInfo & info = (*allocList)[i];
				char fileStr [128];
				strncpy(fileStr, info.file, 128);
				char * fileStrPtr = (char*) strrchr(fileStr, (int)'\\');
				if(fileStrPtr) fileStrPtr++;
				else		   fileStrPtr = (char*) fileStr;

				int fileLength = strlen(fileStrPtr);
				if(maxFileLength < fileLength)
					maxFileLength = fileLength;

				char numStr [64];
				sprintf(numStr, "%d", info.line);

				int lineNumLength = strlen(numStr);
				if(maxLineNumLength < lineNumLength)
					maxLineNumLength = lineNumLength;
			}

			// print header info
			{
				char tempStr [256];

				sprintf(tempStr, "__LINE__");
				strcpy(buf, tempStr);

				sprintf(tempStr, "__FILE__");
				strcatAfter(buf, tempStr, 5+maxLineNumLength+1 + 3);

				sprintf(tempStr, "__SIZE__");
				strcatAfter(buf, tempStr, 5+maxLineNumLength+1 + 3+maxFileLength+1);

				sprintf(tempStr, "__CODE__\n");
				strcatAfter(buf, tempStr, 5+maxLineNumLength+1 + 3+maxFileLength+1 + 32);

				OutputDebugStr(buf);
			}

			// loop through and print all relevant info about each memory leak in chronological order:
			for(unsigned int i = 0 ; i < allocList->size() ; i++)
			{
				AllocInfo & info = (*allocList)[i];
				char fileStr [128];
				strncpy(fileStr, info.file, 128);
				char * fileStrPtr = (char*) strrchr(fileStr, (int)'\\');
				if(fileStrPtr) fileStrPtr++;
				else		   fileStrPtr = (char*) fileStr;

				bool alreadyDone = false;
				int bytes = 0;
				int times = 0;
				int lastSize = 0;

				// for each allocList entry remaining (i.e. for each memory leak),
				for(unsigned int j = 0 ; j < allocList->size() ; j++)
				{
					AllocInfo & info2 = (*allocList)[j];
					if(info.line == info2.line
					&& _stricmp(info.file, info2.file) == 0) // if they're the same entries,
					{
						if(j < i)
						{
							alreadyDone = true; // already included this entry in a multi-entry output
							break;
						}
						else
						{
							bytes += info2.size; // increment number of entries this output corresponds with
							times++;
							lastSize = info2.size;
						}
					}
				}

				if(alreadyDone)
					continue; // already included this entry in a multi-entry output

				char tempStr [256];

				sprintf(tempStr, "line %d ", info.line);
				strcpy(buf, tempStr);

				sprintf(tempStr, "of %s ", fileStrPtr);
				strcatAfter(buf, tempStr, 5+maxLineNumLength+1);

				char num2 [32];
				if(bytes <= 1024)             sprintf(num2, "%d byte%s", bytes, bytes == 1?"":"s");
				else if(bytes <= 1024 * 1024) sprintf(num2, "%g KB", bytes / 1024.0f);
				else                          sprintf(num2, "%g MB", bytes / (1024.0f * 1024.0f));

				const int bytesPer = bytes / times;
				if(bytesPer == lastSize && times > 1)
				{
					char num1 [32];
					if(bytesPer <= 1024)             sprintf(num1, "%d byte%s", bytesPer, bytesPer == 1?"":"s");
					else if(bytesPer <= 1024 * 1024) sprintf(num1, "%g KB", bytesPer / 1024.0f);
					else                             sprintf(num1, "%g MB", bytesPer / (1024.0f * 1024.0f));

					sprintf(tempStr, "%s x%d = %s", num1, times, num2);
					strcatAfter(buf, tempStr, 5+maxLineNumLength+1 + 3+maxFileLength+1);
				}
				else
				{
					if(times > 1)
					{
						sprintf(tempStr, "(varying) x%d = %s", times, num2);
						strcatAfter(buf, tempStr, 5+maxLineNumLength+1 + 3+maxFileLength+1);
					}
					else
					{
						sprintf(tempStr, "%s x1", num2);
						strcatAfter(buf, tempStr, 5+maxLineNumLength+1 + 3+maxFileLength+1);
					}
				}

				// try to get the offending line of code from the file
				{
					FILE* pFile = fopen(info.file, "r");
					if(pFile)
					{
						char str [1024*8];
						str[0] = '\0';
						for(unsigned int i = 0 ; i < info.line && !ferror(pFile) && !feof(pFile); i++)
							fgets(str, 1024*8-1, pFile); 
						str[1024*8-1] = '\0';

						if(!(ferror(pFile) || feof(pFile)))
						{
							// trim leading whitespace
							char* strPtr = str;
							while(strPtr[0] == ' ' || strPtr[0] == '\t' || strPtr[0] == '\n')
								strPtr++;
_lookForSemi:
							// cut off the line at the first not-in-quotes semicolon
							char* strPtr2 = strPtr;
							int quotes = 0;
							bool hitSemicolon = false;
							while(strPtr2[0])
							{
								if((!quotes || quotes == 1) && strPtr2[0] == '\'')
									if(!quotes) quotes = 1;
									else        quotes = 0;
								if((!quotes || quotes == 2) && strPtr2[0] == '\"')
									if(!quotes) quotes = 2;
									else        quotes = 0;
								if(!quotes)
								{
									if(strPtr2[0] == ';')
										++strPtr2, strPtr2[0] = '\0', hitSemicolon = true;
									else if(strPtr2[0] == '/' && strPtr2[1] == '/')
										strPtr2[0] = '\0';
									else strPtr2++;
								}
								else strPtr2++;
							}

							// remove trailing whitespace
							while(strPtr2 != strPtr2 && (strPtr2[-1] == ' ' || strPtr2[-1] == '\t' || strPtr2[-1] == '\n'))
							{
								strPtr2--;
								strPtr2[0] = '\0';
							}

							// if we haven't hit a semicolon yet, combine with next line until we do
							if(strlen(strPtr) < 1024)
							if(!hitSemicolon && !ferror(pFile) && !feof(pFile))
							{
								char str2 [1024*4];
								str2[0] = '\0';
								fgets(str2, 1024*4-1, pFile);
								str2[1024*4-1] = '\0';

								char* strPtr3 = str2;
								while(strPtr3[0] == ' ' || strPtr3[0] == '\t')
									strPtr3++;
								if(strPtr2[-1] != ' ' && strPtr2[-1] != '\t' && strPtr2[-1] != '\n')
									strcat(strPtr, " ");
								strcat(strPtr, strPtr3);
								goto _lookForSemi;
							}

							// change all tabs and newlines to spaces for output
							strPtr2 = strPtr;
							while(strPtr2[0])
							{
								if(strPtr2[0] == '\n' || strPtr2[0] == '\t')
									strPtr2[0] = ' ';
								strPtr2++;
							}

							// take a wild guess at the line's validity
							if(strstr(strPtr, "new") || strstr(strPtr, "NEW") || strstr(strPtr, "New")
							|| strstr(strPtr, "alloc") || strstr(strPtr, "ALLOC")
							|| strstr(strPtr, "make") || strstr(strPtr, "Make")
							|| strstr(strPtr, "reate") || strstr(strPtr, "REATE"))
							{
								strcatAfter(buf, strPtr, 5+maxLineNumLength+1 + 3+maxFileLength+1 + 32);
							}
							else
							{
								// unsure
								strcatAfter(buf, "(?) ", 5+maxLineNumLength+1 + 3+maxFileLength+1 + 32);
								strcat(buf, strPtr);
							}
						}

						fclose(pFile);
					}
					strcat(buf, "\n");
				}

				OutputDebugStr(buf);

				{
					char fileStr [256];
/*					char dirStr [256]; // gets fully-qualified name
					if(info.file[0] == '.')
					{
						GetCurrentDirectory(256, dirStr);
						strcpy(fileStr, dirStr);
						strcat(fileStr, info.file + 1);
					}
					else*/
					{
						strcpy(fileStr, info.file);
					}
					debug.FixPath(fileStr);
					debug.Output("%s(%d) : leaked %d bytes", fileStr, info.line, bytes);
					if(totalSize == 0)
						::OutputDebugString("   (double-click here to see the code)");
					debug.OutputNewline();
				}

				totalSize += bytes;
			}
		}

			sprintf(buf, "\nTotal Unfreed: %I64d byte%s", totalSize, totalSize == 1?"":"s"),		OutputDebugStr(buf);
		if(totalSize > 1024)
		{
			if(totalSize <= 1024 * 1024)
				sprintf(buf, " = %g KB", totalSize / 1024.0f),							OutputDebugStr(buf);
			else
				sprintf(buf, " = %g MB", totalSize / (1024.0f * 1024.0f)),				OutputDebugStr(buf);
		}
		sprintf(buf, "\n\n-----------------------------------------------------------\n");	OutputDebugStr(buf);
		sprintf(buf,     " End Dump of Memory Leaks\n");									OutputDebugStr(buf);
		sprintf(buf,     "-----------------------------------------------------------\n");	OutputDebugStr(buf);

		// bring up the file on closing the app
		if(!wasEmpty)
			debug.RunAnyFile(CLEAN_MEM_LOG_FILE, false);

		checkLeakScope = false;
		customFreeEnabled = false;

		_fcloseall();
	};

	// adds an allocated memory entry
	// TODO: if memory has already been allocated on the same line of the same file,
	//       this should just increase the 'size' value of the existing track
	//       as opposed to adding a whole new track and summing up their sizes later which is done now
	void AddTrack(unsigned long addr,  unsigned long asize,  const char *fname, unsigned long lnum, bool array)
	{
		#ifdef DEBUG_EXTENDED_CHECK_ALLOCS
			try {
				AllocInfo fake_info (addr);
				AllocList::iterator i2 = std::find(allocList->begin(), allocList->end(), fake_info);
 				assertMsg(i2 == allocList->end() || (*i2).address != addr, "something is already allocated at this address!");
			} catch (...) {
				{_asm int 3} // error-checking routine is crashing!
			}
		#endif

		if(!checkLeakScope)
			BeginCheckingUnfreed();

		AllocInfo info;

		info.address = addr;
		info.file = fname;
		info.line = (unsigned short)lnum;
		info.size = asize;
		info.isArray = array;
		customFreeEnabled = false;
		allocList->push_back(info);
		customFreeEnabled = true;

#ifdef DEBUG_OUTPUT_ALL_ALLOCS
		try {
		totalAllocated += asize;
#	ifdef DEBUG_OUTPUT_ALL_ALLOCS_ONLY_ON_CTRL
		if(GetAsyncKeyState(VK_CONTROL) & 0x8000)
#	endif
		{
			char tempStr [512], buf [2048];

			sprintf(tempStr, "ALLOC line %d ", info.line);
			strcpy(buf, tempStr);

			sprintf(tempStr, "of %s ", info.file ? max(info.file, strrchr(info.file, '\\')+1) : "(unknown)");
			strcat(buf, tempStr);

			sprintf(tempStr, "%d bytes / %d total\n", info.size, totalAllocated);
			strcat(buf, tempStr);

			OutputDebugStr(buf);
		}
		} catch (...) {_asm int 3} // error-check the error-check
#		endif
	};

	void HandleMismatch (AllocInfo info, unsigned long addr, bool array)
	{
		char newCode [2048];
		char deleteCode [2048];
		char newFile [128];
		char * newFilePtr = (char*)newFile;
		char deleteFile [128];
		char * deleteFilePtr = (char*)deleteFile;
		newCode[0] = '\0';
		deleteCode[0] = '\0';
		newFile[0] = '\0';
		deleteFile[0] = '\0';

		FILE* pFile = NULL;

		if(info.file)
		{
			pFile = fopen(info.file, "r");
			if(pFile)
			{
				for(unsigned int i = 0 ; i < info.line && !ferror(pFile) && !feof(pFile); i++)
					fgets(newCode, 2048-1, pFile); 
				newCode[2048-1] = '\0';
			}
			strncpy(newFile, info.file, 127);
			newFilePtr = (char*) strrchr(newFile, (int)'\\');
			if(newFilePtr) newFilePtr++;
			else		   newFilePtr = (char*) newFile;
		}

		if(allocDeleteFile)
		{
			pFile = fopen(allocDeleteFile, "r");
			if(pFile)
			{
				for(int i = 0 ; i < allocDeleteLine && !ferror(pFile) && !feof(pFile); i++)
					fgets(deleteCode, 2048-1, pFile); 
				deleteCode[2048-1] = '\0';
			}
			strncpy(deleteFile, allocDeleteFile, 127);
			deleteFilePtr = (char*) strrchr(deleteFile, (int)'\\');
			if(deleteFilePtr) deleteFilePtr++;
			else		   deleteFilePtr = (char*) deleteFile;
		}

		if(info.file && allocDeleteFile)
		{
			debug.AssertBox("Mismatch of new%s with delete%s!\n\nnew%s on line %d of %s:\n%sdelete%s on line %d of %s:\n%s", info.isArray ? "[]" : "", array ? "[]" : "", info.isArray ? "[]" : "", info.line, newFilePtr, newCode, array ? "[]" : "", allocDeleteLine, deleteFilePtr, deleteCode);
		}
		else if(info.file)
		{
			debug.AssertBox("Mismatch of new%s with delete%s!\n\nnew%s on line %d of %s:\n%smismatches with a delete%s.", info.isArray ? "[]" : "", array ? "[]" : "", info.isArray ? "[]" : "", info.line, newFilePtr, newCode, array ? "[]" : "");
		}
		else if(allocDeleteFile)
		{
			debug.AssertBox("Mismatch of new%s with delete%s!\n\nnew%s delete%s on line %d of %s:\n%s", info.isArray ? "[]" : "", array ? "[]" : "", info.isArray ? "[]" : "", array ? "[]" : "", allocDeleteLine, deleteFilePtr, deleteCode);
		}
		else
		{
			debug.AssertBox("Mismatch of new%s with delete%s!", info.isArray ? "[]" : "", array ? "[]" : "");
		}
	}


	// removes an allocated memory entry (when mem is de-allocated)
	bool RemoveTrack(unsigned long addr, bool array)
	{
		if(allocList && !allocList->empty())
		{
			AllocInfo fake_info (addr);
			AllocList::iterator i = std::find(allocList->begin(), allocList->end(), fake_info);
 			if(i != allocList->end())
			{
				AllocInfo & info = *i;
 				if(info.address == addr)
				{
					if(info.isArray != array)
						HandleMismatch(info, addr, array);

			#ifdef DEBUG_OUTPUT_ALL_ALLOCS
					try {
					totalAllocated -= info.size;
			#	ifdef DEBUG_OUTPUT_ALL_ALLOCS_ONLY_ON_CTRL
					if(GetAsyncKeyState(VK_CONTROL) & 0x8000)
			#	endif
					{
						char tempStr [512], buf [2048];

						sprintf(tempStr, "FREE at line %d ", allocDeleteLine);
						strcpy(buf, tempStr);

						sprintf(tempStr, "of %s ", allocDeleteFile ? max(allocDeleteFile, strrchr(allocDeleteFile, '\\')+1) : "(unknown)");
						strcat(buf, tempStr);

						sprintf(tempStr, "-%d bytes / %d total\n", info.size, totalAllocated);
						strcat(buf, tempStr);

						OutputDebugStr(buf);
					}
					} catch (...) {_asm int 3} // error-check the error-check
			#		endif

					customFreeEnabled = false;
					allocList->erase(i);
					customFreeEnabled = true;

	#				ifdef DEBUG_EXTENDED_CHECK_ALLOCS
						try {
							AllocList::iterator i2 = std::find(allocList->begin(), allocList->end(), fake_info);
 							assertMsg(i2 == allocList->end() || (*i2).address != addr, "something is still already allocated at this address!");
						} catch (...) {
							{_asm int 3} // error-checking routine is crashing!
						}
	#				endif

					return true;
				}
			}
		}

		return false;

//		if(checkLeakScope)
//		{
//			throw "deleting unallocated";
//			char buf [2048];
//			sprintf(buf, "Invalid Removal: address %d, via delete%s\n", addr, array?"[]":"");
//			debug.Output(buf);
//		}
	};

	void * __malloc(size_t size, const char *file, int line)
	{
		void * ptr;

		#ifndef CLEAN_MEM_NO_VERIFY
		try
		{
		#endif

		ptr = (void *)malloc(size);

		if(isDebug)
			AddTrack((unsigned long)ptr, size, file, line, false);

		#ifndef CLEAN_MEM_NO_VERIFY
		}
		catch(...)
		{
			if(file && isDebug)
			{
				char fileStr [128];
				strncpy(fileStr, file, 127);
				char * fileStrPtr = (char*) strrchr(fileStr, (int)'\\');
				if(fileStrPtr) fileStrPtr++;
				else		   fileStrPtr = (char*) fileStr;

				debug.AssertBox("NEW of %d bytes FAILED on line %d of %s\n", size, line, fileStrPtr);
			}
		}
		#endif

		return ptr;
	}

	void __free(void * mem)
	{
		if(!mem)
		{
			allocDeleteFile = NULL;
			allocDeleteLine = -1;
			return;
		}

		#ifndef CLEAN_MEM_NO_VERIFY
		try
		{
		#endif

#ifdef DEBUG_EXTENDED_CHECK_ALLOCS
		if(checkLeakScope && !RemoveTrack((unsigned long)mem, false) && allocDeleteFile)
			throw "deleting unallocated";
#else
		if(checkLeakScope)
			RemoveTrack((unsigned long)mem, false);
#endif

		free(mem);

		#ifndef CLEAN_MEM_NO_VERIFY
		}
		catch(...)
		{
			if(allocDeleteFile && checkLeakScope)
			{
				char fileStr [128];
				strncpy(fileStr, allocDeleteFile, 127);
				char * fileStrPtr = (char*) strrchr(fileStr, (int)'\\');
				if(fileStrPtr) fileStrPtr++;
				else		   fileStrPtr = (char*) fileStr;

				debug.AssertBox("DELETE of 0x%x FAILED on line %d of %s\n", (unsigned long)mem, allocDeleteLine, fileStrPtr);
			}
			else
			{
				free(mem);
			}
		}
		#endif

		allocDeleteFile = NULL;
		allocDeleteLine = -1;
	}

	// overloaded new
	void * operator new(size_t size, const char *file, int line)
	{
		return __malloc(size,file,line);
	};

	// overloaded new[]
	void * operator new[](size_t size, const char *file, int line)
	{

		void * ptr;

		#ifndef CLEAN_MEM_NO_VERIFY
		try
		{
		#endif

		ptr = (void *)malloc(size);

		if(isDebug)
			AddTrack((unsigned long)ptr, size, file, line, true);

		#ifndef CLEAN_MEM_NO_VERIFY
		}
		catch(...)
		{
			if(file && isDebug)
			{
				char fileStr [128];
				strncpy(fileStr, file, 127);
				char * fileStrPtr = (char*) strrchr(fileStr, (int)'\\');
				if(fileStrPtr) fileStrPtr++;
				else		   fileStrPtr = (char*) fileStr;

				debug.AssertBox("NEW[] of %d bytes FAILED on line %d of %s\n", size, line, fileStrPtr);
			}
		}
		#endif

		return ptr;
	};

	// overloaded delete
	void operator delete(void * mem)
	{
		if(customFreeEnabled)
			__free(mem);
		else
			free(mem), customFreeEnabled = true;
	};

	// overloaded delete
	void operator delete[](void * mem)
	{
		if(!customFreeEnabled)
		{
			free(mem);
			customFreeEnabled = true;
			return;
		}

		if(!mem)
		{
			allocDeleteFile = NULL;
			allocDeleteLine = -1;
			return;
		}

		#ifndef CLEAN_MEM_NO_VERIFY
		try
		{
		#endif

#ifdef DEBUG_EXTENDED_CHECK_ALLOCS
		if(checkLeakScope && !RemoveTrack((unsigned long)mem, true) && allocDeleteFile)
			throw "deleting unallocated";
#else
		if(checkLeakScope)
			RemoveTrack((unsigned long)mem, true);
#endif

		free(mem);

		#ifndef CLEAN_MEM_NO_VERIFY
		}
		catch(...)
		{
			if(allocDeleteFile && checkLeakScope)
			{
				char fileStr [128];
				strncpy(fileStr, allocDeleteFile, 127);
				char * fileStrPtr = (char*) strrchr(fileStr, (int)'\\');
				if(fileStrPtr) fileStrPtr++;
				else		   fileStrPtr = (char*) fileStr;

				debug.AssertBox("DELETE[] of 0x%x FAILED on line %d of %s\n", (unsigned long)mem, allocDeleteLine, fileStrPtr);
			}
			else
			{
				free(mem);
			}
///			Engine.ErrorDialog(tempStr);
		}
		#endif

		allocDeleteFile = NULL;
		allocDeleteLine = -1;
	};

#endif // !CLEANMEM_DISABLE

