//
// GameX - Main-Loop-related GameXClass Class Code (including WinMain and WinProc (WindowFunc()))
// 
// Copyright (C) 2002 Rama C. Hoetzlein
// 
// This software is released under the GameX GNU GPL 
// Open Source License. See the GameX documentation included
// with this source code for terms of modification, 
// distribution and re-release. 
// 
// this is a version of GameX that has been modified (by Justin Pease)


// KNOWN ISSUES:

// In windowed mode, if you click out of a game, then click once on the game's title bar to get back into the game,
// Windows forgets to visibly activate the game window even though GameX still gets the activation message,
// resulting in the game playing sort-of in the background until it's clicked on again.
// May not be universal across computers or specific to GameX, however.

// Direct3D
#ifndef DIRECT3D_VERSION
#define DIRECT3D_VERSION 0x0700 // specify version 7.0 of Direct3D -- the last version that's compatible with DirectDraw
#include <d3d.h>				// note: for newer versions of D3D we would need to 
//#include <d3dtypes.h>			//		 include different files, like d3d8.h
//#include "d3dx.h"
#endif


#include "gamex-win-dx.h"
#include "gamex-win-dx-timer.inl"
#include "gamex-image.h"
#include "gamex-text.h"


long z_lastActiveTime = 0;

// WinMain and WindowFunc functions supplied by GameXClass
LRESULT CALLBACK WindowFunc (HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	return GameX.z_HandleMessage(hWnd, message, wParam, lParam);
}

LRESULT GameXClass::z_HandleMessage (HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	switch (message)
	{
	case WM_MOVE: // window movement detection (non-windowed mode)
		z_SetPosition (LOWORD(lParam), HIWORD(lParam));
		z_SetMouseStop(1); // so window movement doesn't register as mouse movement to the game
		break;
	case WM_CHAR: // buffered keyboard input detection, for GetKey()
		z_AddBufferedKeyPress ((TCHAR) wParam); // add key char to buffer
		return 0;
	case WM_SYSKEYDOWN: // detects pressing of system keys such as ALT
		if(wParam == VK_MENU)
			return 0; // prevent pressing the alt key on its own from pausing/freezing the game
		break;		  // (returning 0 stops Windows from taking action, breaking lets it do what it wants)
	case WM_SYSCOMMAND: // detects system commands such as closing the window, etc.
		switch(wParam & 0xFFF0)
		{
			case SC_CLOSE:
				if(!z_IsQuitDialogOpen())
					Quit();
				return 0;
			case SC_MONITORPOWER:
				if(IsFullScreen())
					return 0;
			case SC_KEYMENU:
			case SC_PREVWINDOW:
				return 0; // prevents ALT-anykey from blinking the window and beeping
						  // and prevents ALT-F6 from pausing the game
						  // but allows ALT-F4 to continue functioning as quit button
		}
		// no break
	case WM_ACTIVATEAPP: // game switched to
		// no break 	  // or
	case WM_ENTERMENULOOP: // menu opened
		if(z_CanDraw() && z_CanAutoPause() && !z_IsQuitDialogOpen())
		{
			SetWarningOnDraw(false);
			z_GameDraw();
															// to indicate game being paused because it is in the background 
			z_win_alternator = 0;						// or has a menu or system command being accessed,
			if(!z_isdrawing)
			z_isdrawing = true, SpecialPauseFade(0.125f,0.25f,0,0,32,64,64,128,160), z_isdrawing = false; // partially fade the screen (blue-ish)
		}
		UpdateView();
		break;
	case WM_KILLFOCUS: // window deactivated
		if(z_win_active || IsWindowActive())
		{
///			if(z_win_fullscreen && z_win_fake_fullscreen)
///				ShowWindow(z_win_hwnd, FALSE);
			ClipCursor(NULL);

			z_SetActivation(false);
			if(z_CanAutoPause())
			{
				PauseMusic();
				char title_str [256];
				sprintf(title_str, "Paused: %s - " GAMEX_VERSION_STRING, GetName());
///				sprintf(title_str, "Paused: %s", GetName());
				SetWindowText(GetWindow(),title_str);
			}
//			if(z_CanHideCursor())
//				z_UnhideCursor();
			z_SetMouseMove(z_mouse_x, z_mouse_y);
			if(z_CanAutoPause()) 
				z_SetMouseStop(5); // skip 5 mouse input loops, the mouse should be ready for input by then
		}
		break;
	case WM_SETFOCUS: // window activated
		if(z_win_fullscreen || z_win_request_keep_cursor_in_window)
		{
			RECT rect = {z_win_xpos+z_win_xoffset, z_win_ypos+z_win_yoffset, z_win_xpos+z_win_xoffset+z_win_realx, z_win_ypos+z_win_yoffset+z_win_realy} ;
			ClipCursor(&rect);
		}
		if(!z_win_active && z_win_realy>1)
		{
///			if(z_win_fullscreen && z_win_fake_fullscreen)
///				ShowWindow(z_win_hwnd, TRUE);

			z_SetActivation (true);
			if(z_CanAutoPause())
			{
				ResumeMusic();
				char title_str [256];
				sprintf(title_str, "%s - " GAMEX_VERSION_STRING, GetName());
///				sprintf(title_str, "%s", GetName());
				SetWindowText(GetWindow(),title_str);
			}
//			if(z_CanHideCursor())
//				z_HideCursor();
			z_SetMouseMove(z_mouse_x, z_mouse_y);
		}
		break;
	case WM_PAINT: // redraw of game screen requested
		UpdateView(); // (in addition to the frequent timed updates, in case game is suspended)
		break;
	case WM_CLOSE: // close box or command chosen
		if(z_CheckQuit())
		{
			z_SetStatus(GAMEX_SHUTTINGDOWN);
			DestroyWindow (hWnd); // quit the game by activating WM_DESTROY case next
		}
		else
			return 0; // user decided against quitting, continue game
		break;
	case WM_DESTROY:
		ClipCursor(NULL);
		PostQuitMessage(0); // post the message that actually quits the application/game
		break;

	// these are used to check to make sure DirectInput's mouse presses are within the window, and to add mousedown/up events
	case WM_LBUTTONDOWN:	z_SetMousePress(MOUSE_LEFT)/*,   z_AddMouseDownEvent(MOUSE_LEFT)*/;			break;
	case WM_MBUTTONDOWN:	z_SetMousePress(MOUSE_MIDDLE)/*, z_AddMouseDownEvent(MOUSE_MIDDLE)*/;		break;
	case WM_RBUTTONDOWN:	z_SetMousePress(MOUSE_RIGHT)/*,  z_AddMouseDownEvent(MOUSE_RIGHT)*/;		break;
	//case WM_LBUTTONUP:		z_AddMouseUpEvent(MOUSE_LEFT);			return 0;
	//case WM_MBUTTONUP:		z_AddMouseUpEvent(MOUSE_MIDDLE);		return 0;
	//case WM_RBUTTONUP:		z_AddMouseUpEvent(MOUSE_RIGHT);			return 0;

	// these are used to allow double-click detection (only works if CS_DBLCLKS is set on our WNDCLASS.style)
	case WM_LBUTTONDBLCLK:	z_SetMouseDoubleClick(MOUSE_LEFT);		break;
	case WM_MBUTTONDBLCLK:	z_SetMouseDoubleClick(MOUSE_MIDDLE);	break;
	case WM_RBUTTONDBLCLK:	z_SetMouseDoubleClick(MOUSE_RIGHT); 	break;

///	case WM_KEYDOWN:		z_AddKeyDownEvent((int)wParam);		return 0;
	case WM_KEYUP:			///z_AddKeyUpEvent((int)wParam);
		if(wParam == VK_SNAPSHOT) { // ("Prnt Scrn" button released)
			if(z_CanShowRunInfo()) {SetWarningOnDraw(false); z_GameDraw();} // clear debug info if it's on

			char name [MAX_PATH];
			FILE * file = NULL;
			for(int i = 1 ; ; i++)
			{
				sprintf(name, "screenshot%d.bmp", i);
				if(!debug.DoesFileExist(name))
					break;
			}

			ImageExt ie; ie.Save(name, VIEWPORT); // make taking screenshots easier
			CopyToClipboard(VIEWPORT); // copy back buffer directly to clipboard as well as to file
		}
		break;

	case WM_SIZE: // user resizes the game window in windowed mode
		if(z_GetStatus() == GAMEX_READY)
		{
			if(HIWORD(lParam) < 1)
			{
				PostMessage(hWnd,WM_KILLFOCUS,0,0);
			}
			else
			{
				PostMessage(hWnd,WM_SETFOCUS,0,0);
			}

			if(!z_IsResizeWorkaroundEnabled())
			{
				z_win_realx = LOWORD(lParam);
				z_win_realy = HIWORD(lParam);
				z_SetProportional();
			}
			else
			{
				z_win_xoffset = (LOWORD(lParam) - GetWidth()) / 2;
				z_win_yoffset = (HIWORD(lParam) - GetHeight()) / 2;				
				z_win_realx = GetWidth();
				z_win_realy = GetHeight();
			}

			if(z_win_realy < 1) z_win_realy = 1; // avoids crash
			if(z_win_realx < 1) z_win_realx = 1;
			z_SetMouseStop(1); // so window resize doesn't register as mouse movement to the game

			GameX.ResumeMusic(); // HACK resize movies
		}
		break;
	case WM_DIRECTSHOW_EVENT:
		z_HandleDirectShowEvent();
		break;
	}
	
	return DefWindowProc (hWnd, message, wParam, lParam);		
}

void CallGameCleanup() // a separate function for extra logging information
{
	if(debug.IsRunningDebug())
		debug.Output("Running GameCleanup...\n");

	GameX.z_SetStatus(GAMEX_SHUTTINGDOWN);

	// uninitialize the game
	GameCleanup();

	if(debug.IsRunningDebug())
		debug.Output("GameCleanup Finished.\n");
}

int WINAPI WinMain (HINSTANCE hThisInst, HINSTANCE hPrevInst, LPSTR lpszArgs, int nWinMode)
{
	return GameX.z_StartMain(hThisInst, hPrevInst, lpszArgs, nWinMode);
}

int GameXClass::z_StartMain (HINSTANCE hThisInst, HINSTANCE hPrevInst, LPSTR lpszArgs, int nWinMode)
{
	MSG msg;
	
	z_InitWindowsInstance (hThisInst, nWinMode);
	z_ParseArgs(lpszArgs);

	if(z_win_log_debug)
	{
		debug.OutputTime();
		debug.Output("Running GameInit...\n");
	}

	GameInit(); // defined by game, hopefully calls Initialize()
	atexit(CallGameCleanup); // call GameCleanup() when returning from this static scope while exiting the game

	if(z_GetStatus() != GAMEX_READY)
	{
		z_ReportProblem();
		z_SetStatus(GAMEX_READY); // after temp auto-init
	}

	z_win_init_done = true;
	z_isdrawing = false;

	z_LoadQueuedImages();

	if(z_win_log_debug)
	{
		debug.Output("Graphic/Sound memory in use after GameInit(): %d KB\n", z_GetGraphicMemUsed());
		debug.Output("GameX: Running Game...\n");
	}

	int frame = 0, cycle = 0; // so we can keep track of the frame rate
	z_sleep_record = 0; // to keep track of how much time we spend sleeping

	z_win_global_counter = 0;
	z_win_global_frame_total = 0;
	long long cps = z_win_cps; // cycles per second (a cycle meaning a call to GameRun(), not a CPU cycle)
	long long fps = z_win_fps; // frames per second (a frame meaning a call to GameDraw() and UpdateView())
	long long tpc = GetClocksPerSec() / (cps?cps:1); // time per cycle (number of clock ticks between each game update)
	long long tpf = GetClocksPerSec() / (fps?fps:1); // time per frame (number of clock ticks between each view update)
	bool needUpdate = false; // (start as false because GameRun() should be called first before GameDraw())
	bool update_auto_status = true;
	long long cycleTime = GetClock(); // timer for game update
	long long frameTime = GetClock(); // timer for view update
	long long frameClock = GetClock(); // timer for framerate calculation update
	int leftover = 0;

	z_DisplayLoadingWithValue (0); // clear message

	if(cps == 0) // if frame rate management is turned off, loop stupidly:
	{
		long long count = 0;
		long long lastCount = 0;
		for(;;)
		{
			while(PeekMessage(&msg, NULL, 0, 0, PM_REMOVE))
			{
				if(msg.message == WM_QUIT) goto game_exit_label;
				TranslateMessage(&msg);
				DispatchMessage(&msg);
			}
			if(GetWindow() == GetForegroundWindow())
				z_lastActiveTime = clock();

			count++;

			SetWarningOnDraw(true);
			z_UpdateInput();
			GameRun(); // call GameRun() function, whose contents decide the behaviour of the game
			z_win_global_counter++;
			SetWarningOnDraw(false);

			if(z_CanDraw()) z_GameDraw(); // call z_GameDraw() function, whose contents draw the current game state

							if(z_CanShowRunInfo())
							{
								DrawStatus(update_auto_status);
								if(update_auto_status) update_auto_status = false;
							}

			UpdateView(); // update onscreen view (NOTE: this also waits for VSync unless VIDEO_DISALLOWREFRESHSYNC is set)
			z_win_global_frame_total++;
			z_win_alternator = !z_win_alternator;

			static long long thetime = GetClock(); // calculate the frame and cycle rate every second
			if(GetClock() - thetime > GetClocksPerSec())         // so the game doesn't need to calculate it itself
			{                                        // and can instead call GetCurrentFrameRate().
				thetime = GetClock();
				z_f_rate = (unsigned short)(count - lastCount);
				z_c_rate = (unsigned short)(count - lastCount);
				z_p_idle = (int)(100.0*(double)z_sleep_record/(double)GetClocksPerSec()); // calculate idle time percent (percent of power conserved)
				if(z_p_idle < 0 || z_p_idle > 99) z_p_idle = 99;
				z_sleep_record = 0;
				update_auto_status = true;
				debug.Flush();
				lastCount = count;
			}

		}
	}
	else // if frame rate management is on:
	{
		for(;;)
		{
			if(PeekMessage(&msg, NULL, 0, 0, PM_REMOVE))
			{
				if(msg.message == WM_QUIT) goto game_exit_label;
				TranslateMessage(&msg);
				DispatchMessage(&msg);
			}
			if(!z_win_active && IsWindowActive())
			{
				PostMessage(GetWindow(),WM_SETFOCUS,0,0);
			}
			else if(z_win_active && !IsWindowActive())
			{
				PostMessage(GetWindow(),WM_KILLFOCUS,0,0);
			}
			if(z_win_active)
				z_lastActiveTime = clock();
			if(!z_win_active && z_CanAutoPause()) // if the window isn't active, and we can pause,
			{
				Sleep(10); z_sleep_record += GetClocksPerSec() * 10 / CLOCKS_PER_SEC; // sleep so we don't max out the cpu in the background
				debug.Flush(); // release hold on the debug log file
				cycleTime = GetClock(); // set the timers to the current time
				frameTime = GetClock(); // so we don't skip a ton of frames when resuming
			}
			else
			{
				// AUTOMATIC GAME SPEED AND FRAME RATE MANAGEMENT:
				// This ensures that the game always runs at a given speed (chosen by the game programmer)
				// regardless of how quickly the computer is able to execute the game's drawing/instructions.

				// CPS means Cycles Per Second. Higher CPS means faster gameplay.	GameRun() is called z_win_cps times per second.
				// FPS means Frames Per Second. Higher FPS means smoother gameplay. GameDraw() is called z_win_fps times per second.
				// Attaining the target CPS (game speed) is nearly guaranteed under normal circumstances, (< 10000 cps)
				// whereas attaining the target FPS (game smoothness) isn't always possible, depending on the game and cpu.

				// Warning: the following code is VERY "fragile" due to its time-sensitive nature, so even
				// the most innocent modification is likely to cause subtle problems if not drastic failure.

				// first check to see if the game changed the cps or fps and adjust accordingly
				if(cps != z_win_cps)
				{
					cps = z_win_cps;
					tpc = GetClocksPerSec() / cps;
					if(fps > cps) tpf = tpc;
					else tpf = GetClocksPerSec() / fps;
					update_auto_status = true;

				}
				if(fps != z_win_fps)
				{
					fps = z_win_fps;
					tpf = GetClocksPerSec() / fps;
					tpf = max(tpf,tpc);
					update_auto_status = true;
				}

				// sleep while draw update would exceed game update
				if(!needUpdate) AccurateSleep(tpc+cycleTime-GetClock(), 1);

				long long thisTime = GetClock(); // variable to hold pre-action reference time

				int cpf_counter = 0; // number of times to update the game at this instant,
									// if > 1 then we skip (cpf_counter-1) frames by updating again before drawing them

				// calculate how many times we need to update the game
				while(thisTime-cycleTime > tpc)
				{
					cycleTime += tpc;
					cpf_counter++;
				}
				// limit number of times to the maximum allowable updates per frame
				if(cpf_counter > z_win_max_cpf)
				{
					cpf_counter = z_win_max_cpf;
					frameTime = GetClock();
				}
	/*
				// smooth out delay discrepencies between frames:
				// smoothing method 1:
				int new_todo = cpf_counter;
				int total_todo = new_todo + leftover;
				int trying = total_todo ? 1 + total_todo / 2 : 0;
				leftover = total_todo - trying;
				cpf_counter = trying;
	*/
				// smoothing method 2:
				// (difficult to tell which is better -- method 1 is more correct in terms of achieving exactly the desired rate,
				//	but whatever rate it reaches sometimes appears slightly choppier than the same rate achieved with method 2)
				int ave = (cpf_counter+cpf_counter+cpf_counter + leftover) / 4;
				int total = cpf_counter + leftover;
				cpf_counter = total - ave; // takes into account the integer-divided nature of ave
				leftover = ave;

				if(cpf_counter) { // if we need to update the game state at all since last time:
					SetWarningOnDraw(true);
					while(cpf_counter--) { // update (without drawing) as many times as calculated
						z_UpdateInput();
						GameRun(); // call GameRun() function, whose contents decide the behaviour of the game
						cycle++; // increment counter so we can know how many cps we're really getting
						z_win_global_counter++;
						if(cycleTime > thisTime) { // correct overflow
							cycleTime = thisTime;
							cpf_counter = 0;
						}
					}
					needUpdate = true; // remember that the game screen now needs to be drawn again
					SetWarningOnDraw(false);
				}

				// sleep until the next draw update:
///				Sleep(1); z_sleep_record += GetClocksPerSec() / CLOCKS_PER_SEC;
				AccurateSleep(tpf+frameTime-GetClock(), 1);

				// draw the game as necessary:
				thisTime = GetClock();
				if(thisTime-frameTime > tpf) { // executes at most once per loop
					if(needUpdate)
					{
						if(needUpdate && z_CanDraw())
						{
							// call z_GameDraw() function, whose contents draw the current game state
							z_GameDraw();
							// also draw game status if RUN_AUTOSHOWINFO is enabled:
							if(z_CanShowRunInfo())
							{
								DrawStatus(update_auto_status);
								if(update_auto_status) update_auto_status = false;
							}
						}
						UpdateView(); // update onscreen view
						frame++; // increment counter so we can know how many fps we're really getting
						z_win_global_frame_total++;
						z_win_alternator = !z_win_alternator;
						frameTime = min(frameTime+tpf, thisTime);
						needUpdate = false;
					}
				} 

				long long thetime = GetClock() - frameClock; // calculate the frame and cycle rate every second
				if(thetime >= GetClocksPerSec())         // so the game doesn't need to calculate it itself
				{                                        // and can instead call GetCurrentFrameRate().
					z_f_rate = frame;				
					z_c_rate = cycle;
					z_p_idle = (int)(100.0*(double)z_sleep_record/(double)GetClocksPerSec()); // calculate idle time percent (percent of power conserved)
					if(z_p_idle < 0 || z_p_idle > 99) z_p_idle = 99;
					frame = 0; cycle = 0; z_sleep_record = 0;
					frameClock = GetClock();
					if(frameClock - frameTime > GetClocksPerSec()/2)
						frameTime = frameClock; // prevent long-term frame loss compensation
					update_auto_status = true;
					debug.Flush();
				}
			}
		}
	}

game_exit_label:

	if(z_win_log_debug)
		debug.Output("GameX: Exiting normally...\n");

	return (int)(msg.wParam);
}



void GameXClass::SetGameSpeed (int cps)
{
	if(cps < 1 && cps != 0) cps = 1;
	if(cps > 600) cps = 600;
	z_win_cps = cps;
}

void GameXClass::SetLagLimit (int max_cpf)
{
	if(max_cpf == 0) max_cpf = 0xFFFF; // 0 means no limit (for networked games)
	if(max_cpf<3) max_cpf = 3;
	z_win_max_cpf = max_cpf;
}

void GameXClass::SetForcedFrameRate (int fps)
{
	z_win_fps = fps;
	if(fps>z_win_screenrefresh) z_win_vsync_enabled = false;
}

void GameXClass::ResetForcedFrameRate ()
{
	z_win_fps = -1;
}

static int s_scanlineTarget = 0;
static long long s_lastTime = 0;

int GameXClass::UpdateView ()
{
	if(z_win_status != GAMEX_READY || !z_dd_front_buf)
		return 0;

	// end the scene here if it's still open
//	if(z_draw_render_batch_open)
 //		z_EndRenderBatch();

	z_EmptyVBuffer();

	// DirectX 8.1
	//dd_device->Present (NULL, NULL, z_win_hwnd, NULL);

	// DirectDraw
	HRESULT status; 
	
	if(z_dd_front_buf->IsLost())
	{
		z_dd_front_buf->Restore();

		z_dd_was_lost = 1;

		if(z_win_log_debug)
			if(!z_dd_front_buf->IsLost())
				if(z_win_log_debug)
					debug.Output("GameX: DirectX surfaces were lost, but have now been properly restored.\n");

		if(z_dd_front_buf->IsLost())
		{
			z_LockVBuffer();
			return 0;
		}
		else
		{
			z_dd_main->RestoreAllSurfaces();

			// now we restore the 3D drawing device.
			// for some reason, if we don't do so RIGHT HERE
			// strange and inexplicable image corruption occurs

			z_UnlockBackBuffer();
			if(z_d3d_device != NULL)
			{
				unsigned long temp_long = -1;
				status = z_d3d_device->ValidateDevice(&temp_long);

				if(status == DDERR_SURFACELOST) { // in newer versions of DirectX this is really D3DERR_DEVICELOST, which makes more sense
					z_d3d_device->Release();
					bool success = z_CreateDeviceForSurface(z_d3d_device, z_dd_back_buf); // may fail if z_UnlockBackBuffer() is not done before this
					z_d3d_device->SetRenderState(D3DRENDERSTATE_ALPHABLENDENABLE, TRUE); //Turn on alpha blending

					if(success)
					{
						if(z_win_log_debug)
							debug.Output("GameX: The 3D rendering device has been lost and is now properly restored.\n");
					}
					else
					{
						if(z_win_err_debug)
							debug.Output("GameX: The 3D rendering device has been lost and could NOT be restored.\n");
					}
				}
			}
		}
	}
	
	// if we're still lost, reinit everything related to the display
	if(z_IsBackBufLost())
		z_ReInitScreen();
	
	z_UnlockBackBuffer();

	// use Blt since we might want to scale the viewport to the screen
	// and, if in windowed mode, because we have a clipper set to the window

	// set the source and destination rectangles:
	RECT src_rect = {0, 0, z_win_width, z_win_height} ;
	RECT dst_rect = {z_win_xpos+z_win_xoffset, z_win_ypos+z_win_yoffset, z_win_xpos+z_win_xoffset+z_win_realx, z_win_ypos+z_win_yoffset+z_win_realy} ;

	DWORD scanLineAfter, scanLineBefore;
	HRESULT hr1, hr2;

	const int prevPrio = GetPriorityClass(GetCurrentProcess());
	SetPriorityClass(GetCurrentProcess(), REALTIME_PRIORITY_CLASS); // temporarily increase priority for timing-critical section
	{
		// WARNING: freeze in this block of code and you'll almost certainly have to reboot the computer

		if(z_win_vsync_enabled && z_win_init_done)  // if vsync hasn't been disabled,
			z_SyncWithRefresh(); // wait until we're synchronized with the screen's refresh

		scanLineBefore = 0; hr1 = z_dd_main->GetScanLine(&scanLineBefore); // scanline position (in pixels from top) before blitting

		s_lastTime = GetClock();

		// do the blit:  (note that BltFast skips some functionality we need, like scaling and clipping)
		status = z_dd_front_buf->Blt(&dst_rect, z_dd_back_buf,
									&src_rect,
									DDBLT_WAIT,
									NULL);

		scanLineAfter = 0; hr2 = z_dd_main->GetScanLine(&scanLineAfter); // scanline position (in pixels from top) after blitting
	}
	SetPriorityClass(GetCurrentProcess(), prevPrio); // revert priority class


	const DWORD topHeight = max(0, z_win_ypos+z_win_yoffset);
	const DWORD botHeight = min(z_win_screeny, z_win_realy+z_win_ypos+z_win_yoffset);

	const bool scan1bad = scanLineBefore >= topHeight && scanLineBefore <= botHeight && hr1 != DDERR_VERTICALBLANKINPROGRESS;
	const bool scan2bad = scanLineAfter >= topHeight && scanLineAfter <= botHeight && hr2 != DDERR_VERTICALBLANKINPROGRESS;

	if(scan1bad || scan2bad)
	{
//		if(scan1bad && !scan2bad)
//			debug.Output("!");
//		if(scan2bad && !scan1bad)
//			debug.Output("@");
//		if(scan1bad && scan2bad)
//			debug.Output("#");

		static int count = 0;
		if(scan1bad && (!scan2bad || (scanLineBefore + scanLineAfter > botHeight + topHeight)))
		{
			s_scanlineTarget += 10; // too high
			if(s_scanlineTarget > z_win_screeny+50)
			{
				s_scanlineTarget -= z_win_screeny+50;
//				debug.Output("_v_");
			}
			else
			{
//				debug.Output("%dv",s_scanlineTarget);
//				count++;
//				if(count%15 == 0)
//					debug.Output("\n");
			}
		}
		else
		{
			s_scanlineTarget -= 10; // too low
			if(s_scanlineTarget < 0)
			{
				s_scanlineTarget += z_win_screeny+50;
//				debug.Output("_^_");
			}
			else
			{
//				debug.Output("%d^",s_scanlineTarget);
//				count++;
//				if(count%15 == 0)
//					debug.Output("\n");
			}
		}
	}

	// we might clear the back buffer at this point to avoid visual remnants of previous frames hanging around,
	// but it might be intentional in some circumstances for such remnants to be visible,
	// so, to allow greater freedom for the game developer, the back buffer is not automatically cleared.
	// (and the z-buffer clearing is handled elsewhere, in End3DScene)

	// debug overdraw testing
	if(debug.IsRunningDebug() && GameX.IsKeyDown(KEY_NUMLOCK))
	{
		ClearScreen();
	}

	z_LockVBuffer();

	if(SUCCEEDED(status))
		return 1;
	else
		return 0;
}

// VSync method
void GameXClass::z_SyncWithRefresh ()
{
	// (the reason we don't use WaitForVerticalBlank here
	//  is because it wastes a huge amount of CPU processing time,
	//  and doesn't even always give good results.)
//	z_dd_main->WaitForVerticalBlank(DDWAITVB_BLOCKBEGIN, 0); return;

	const long long elapsedTime = GetClock() - s_lastTime;
	const long long timePerFrame = GetClocksPerSec() / z_win_screenrefresh;
	static const long long leeway = GetClocksPerSec() / 600; // skip this much sleep per frame just in case
	static const long long clocksToTicks = GetClocksPerSec() / CLOCKS_PER_SEC;

	// sleep as many milliseconds as we can without going over
	if(timePerFrame - elapsedTime - leeway > 0)
	{
		long long clocks = timePerFrame - elapsedTime - leeway;
		long long sleeptimer = GetClock();
		int sleepAmount = (int)(clocks / clocksToTicks); // convert to milliseconds
		if(sleepAmount > 0)
		{
			Sleep(sleepAmount); // NOTE: we increased the accuracy of Sleep() during startup to 1 ms (this wouldn't work very well otherwise)
			z_sleep_record += GetClock() - sleeptimer;
		}
	}

	// busy loop until it's safe to draw (which should be very soon)
	static const int scanNearTolerance = 20;
	long long startTime = GetClock();
	do
	{
		static DWORD scanline = 0;
		HRESULT hr = z_dd_main->GetScanLine(&scanline); // scanline position in pixels from top
		if((hr == DDERR_VERTICALBLANKINPROGRESS && s_scanlineTarget >= z_win_screeny) // if the monitor's in vblank and our target is low
		|| abs((int)(scanline - (s_scanlineTarget))) < scanNearTolerance) // or we're close enough to the target
			break; // it's safe to draw now

	} while(GetClock() - startTime < timePerFrame*1); // if it takes more than 1 frame of time, that's much too long, give up on waiting
}


void GameXClass::DrawStatus(bool continuously)
{
	if(!(z_win_fullscreen && tp.text_img != NULL))
	{
		int rateMod = (z_f_rate/((z_win_fullscreen)?4:1));
		if(!continuously && z_f_rate != 0 && (rateMod == 0 || z_win_global_frame_total % rateMod !=0))
			return;
	}

	if(!z_win_request_no_mouse_input && (continuously || (z_f_rate != 0 && z_win_global_frame_total%z_f_rate == 0)))
	{
		z_di_statMouseX = GameX.GetMouseX();
		z_di_statMouseY = GameX.GetMouseY();
	}

	char mouseStr [64];
	if(z_win_request_no_mouse_input) sprintf(mouseStr,"");
	else sprintf(mouseStr," - (%d, %d) mouse",z_di_statMouseX,z_di_statMouseY);

	char str [256];
	sprintf(str,"%d/%d fps - %d/%d cps - %d/%d idle%s",GetCurrentFrameRate(),GetTargetFrameRate(),GetCurrentCycleRate(),GetTargetCycleRate(),GetCurrentIdlePercent(),100,mouseStr);

	if(!z_win_fullscreen)
		SetWindowText(z_win_hwnd,str);
	else
		DrawText(12,z_win_height-24,str, 255,255,255, 16.0f);
}


void GameXClass::z_ParseArgs (char* args)
{
	z_win_commandLineArgs.clear();

	// state
	int quoteType = 0;
	const char* strPtr = (const char*) args;

	// loop through args string, purposely including null terminator
	int len = (int)strlen(args);
	for(int i = 0 ; i <= len ; i++)
	{
		// treat '\' as escape sequence
		if(args[i] == '\\')
		{
			strcpy(args+i, args+i+1);
			i++;
		}
		else
		{
			// detect quotes so as to not split up spaces inside them
			if(args[i] == '\"' && quoteType != 2)
				quoteType = quoteType ? 0 : 1;
			if(args[i] == '\'' && quoteType != 1)
				quoteType = quoteType ? 0 : 2;
		}

		// convert whitespace to string end marks if not inside quotes
		if(args[i] == ' ' || args[i] == '\t' || args[i] == '\n' || args[i] == '\0')
		{
			if(!quoteType || args[i] == '\0')
			{
				args[i] = '\0';
				if(i && args[i-1] != '\0')
					z_win_commandLineArgs.push_back(strPtr); // push back a completed arg token
				strPtr = args + i + 1;
			}
		}
	}
}



long z_gamerand = 1;
long z_drawrand = 2;
bool z_isdrawing = false;
bool z_wasdrawing = false;

short GameXClass::GetRand()
{
	return z_isdrawing ? static_cast<short>(((z_drawrand = z_drawrand * 214013L + 2531011L) >> 16) & 0x7fff)
	                   : static_cast<short>(((z_gamerand = z_gamerand * 214013L + 2531011L) >> 16) & 0x7fff);
}

void GameXClass::RandomizeSeed()
{
	if(z_isdrawing)
		z_drawrand += (long)time(NULL) << 2; // note: it's intentional that the drawrand seed generally won't be set
	else
		z_gamerand += (long)time(NULL) << 2;
}

void GameXClass::SetRandomSeed(int seed)
{
	if(z_isdrawing)
		z_drawrand = seed;
	else
		z_gamerand = seed;
}
